if __name__ == "__main__":
    import code
    import readline
    import rlcompleter

    from .decoder import *
    from .formats import *
    from .instruction_table import *
    from .regs import RISCV_REGS

    sess_vars = globals()
    sess_vars.update(locals())

    readline.set_completer(rlcompleter.Completer(sess_vars).complete)
    readline.set_completer(rlcompleter.Completer(sess_vars).complete)
    readline.parse_and_bind("tab: complete")
    code.InteractiveConsole(sess_vars).interact(
        banner="Interactive decoding session started...", exitmsg="Closing..."
    )



"""
RiscEmu (c) 2021 Anton Lydike

SPDX-License-Identifier: MIT
"""
import re
from io import IOBase
from typing import Dict, Tuple, Iterable, Callable, List, TextIO

from .assembler import MemorySectionType, ParseContext, AssemblerDirectives
from .colors import FMT_PARSE
from .helpers import Peekable
from .tokenizer import Token, TokenType, tokenize
from .core import (
    Program,
    T_ParserOpts,
    ProgramLoader,
    SimpleInstruction,
    ParseException,
)


def parse_instruction(token: Token, args: Tuple[str], context: ParseContext):
    if context.section is None:
        context.new_section(".text", MemorySectionType.Instructions)
    if context.section.type != MemorySectionType.Instructions:
        raise ParseException(
            "{} {} encountered in invalid context: {}".format(token, args, context)
        )
    ins = SimpleInstruction(
        token.value, args, context.context, context.current_address()
    )
    context.section.data.append(ins)


def parse_label(token: Token, args: Tuple[str], context: ParseContext):
    name = token.value[:-1]
    if re.match(r"^\d+$", name):
        # relative label:
        context.context.numbered_labels[name].append(context.current_address())
    else:
        if name in context.context.labels:
            print(FMT_PARSE + "Warn: Symbol {} defined twice!".format(name))
        context.add_label(name, context.current_address(), is_relative=True)


PARSERS: Dict[TokenType, Callable[[Token, Tuple[str], ParseContext], None]] = {
    TokenType.PSEUDO_OP: AssemblerDirectives.handle_instruction,
    TokenType.LABEL: parse_label,
    TokenType.INSTRUCTION_NAME: parse_instruction,
}


def parse_tokens(name: str, tokens_iter: Iterable[Token]) -> Program:
    """
    Convert a token stream into a parsed program
    :param name: the programs name
    :param tokens_iter: the programs content, tokenized
    :return: a parsed program
    """
    context = ParseContext(name)

    for token, args in composite_tokenizer(Peekable[Token](tokens_iter)):
        if token.type not in PARSERS:
            raise ParseException("Unexpected token type: {}, {}".format(token, args))
        PARSERS[token.type](token, args, context)

    return context.finalize()


def composite_tokenizer(
    tokens_iter: Iterable[Token],
) -> Iterable[Tuple[Token, Tuple[str]]]:
    """
    Convert an iterator over tokens into an iterator over tuples: (token, list(token))

    The first token is either a pseudo_op, label, or instruction name. The token list are all remaining tokens before
    a newline is encountered
    :param tokens_iter: An iterator over tokens
    :return: An iterator over a slightly more structured representation of the tokens
    """
    tokens: Peekable[Token] = Peekable[Token](tokens_iter)

    while not tokens.is_empty():
        token = next(tokens)
        if token.type in (
            TokenType.PSEUDO_OP,
            TokenType.LABEL,
            TokenType.INSTRUCTION_NAME,
        ):
            yield token, tuple(take_arguments(tokens))


def take_arguments(tokens: Peekable[Token]) -> Iterable[str]:
    """
    Consumes (argument comma)* and yields argument.value until newline is reached
    If an argument is not followed by either a newline or a comma, a parse exception is raised
    The newline at the end is consumed
    :param tokens: A Peekable iterator over some Tokens
    """
    while True:
        if tokens.peek().type == TokenType.ARGUMENT:
            yield next(tokens).value
        elif tokens.peek().type == TokenType.NEWLINE:
            next(tokens)
            break
        elif tokens.peek().type == TokenType.COMMA:
            next(tokens)
        else:
            break

        # raise ParseException("Expected newline, instead got {}".format(tokens.peek()))


class AssemblyFileLoader(ProgramLoader):
    """
    This class loads assembly files written by hand. It understands some assembler
    directives and supports most pseudo instructions. It does very little verification
    of source correctness.

    It also supports numbered jump targets and properly supports local and global scope
    (globl assembly directive)


    The AssemblyFileLoader loads .asm and .s files by default, and acts as a weak
    fallback to all other filetypes.
    """

    is_binary = False

    source: TextIO

    def parse(self) -> Program:
        with self.source as f:
            return parse_tokens(self.filename, tokenize(f))

    @classmethod
    def can_parse(cls, source_name: str) -> float:
        """
        Parse a string assembly file.
        It also acts as a weak fallback if no other loaders want to take the file.
        """
        # gcc recognizes these line endings as assembly. So we will do too.
        if source_name == "-" or source_name.split(".")[-1].lower() in ("asm", "s"):
            return 1
        return 0.01

    @classmethod
    def get_options(cls, argv: List[str]) -> Tuple[List[str], T_ParserOpts]:
        return argv, {}



from typing import Union

from .instruction_set import InstructionSet, Instruction
from ..core import BaseFloat, Int32, Float32


class RV_Debug(InstructionSet):
    def instruction_print(self, ins: Instruction):
        reg = ins.get_reg(0)
        if len(ins.args) == 2:
            msg = ins.args[1]
            print(msg.format(reg, self.regs.get(reg)))
        else:
            print("register {} contains value {}".format(reg, self.regs.get(reg)))

    def instruction_printf(self, ins: Instruction):
        fmt_str = ins.args[0]
        regs = tuple(self.smart_get_reg(x) for x in ins.args[1:])
        print(fmt_str.format(*regs))

    def instruction_print_float(self, ins: Instruction):
        reg = ins.get_reg(0)
        print("register {} contains value {}".format(reg, self.regs.get_f(reg).value))

    def instruction_print_float_s(self, ins: Instruction):
        reg = ins.get_reg(0)
        print(
            "register {} contains value {}".format(
                reg, Float32.bitcast(self.regs.get_f(reg)).value
            )
        )

    def instruction_print_uint(self, ins: Instruction):
        reg = ins.get_reg(0)
        print(
            "register {} contains value {}".format(
                reg, self.regs.get(reg).unsigned_value
            )
        )

    def instruction_print_hex(self, ins: Instruction):
        reg = ins.get_reg(0)
        print(
            "register {} contains value {}".format(reg, hex(self.regs.get(reg).value))
        )

    def instruction_print_uhex(self, ins: Instruction):
        reg = ins.get_reg(0)
        print(
            "register {} contains value {}".format(
                reg, hex(self.regs.get(reg).unsigned_value)
            )
        )

    def smart_get_reg(self, reg_name: str) -> Union[Int32, BaseFloat]:
        if reg_name[0] == "f":
            return self.regs.get_f(reg_name)
        return self.regs.get(reg_name)



"""
RiscEmu (c) 2021 Anton Lydike

SPDX-License-Identifier: MIT
"""

from .instruction_set import *
from riscemu.core.exceptions import INS_NOT_IMPLEMENTED


class RV32M(InstructionSet):
    """
    The RV32M Instruction set, containing multiplication and division instructions
    """

    def instruction_mul(self, ins: "Instruction"):
        rd, rs1, rs2 = self.parse_rd_rs_rs(ins)
        self.regs.set(rd, rs1 * rs2)

    def instruction_mulh(self, ins: "Instruction"):
        rd, rs1, rs2 = self.parse_rd_rs_rs(ins)
        self.regs.set(rd, (rs1 * rs2) >> 32)

    def instruction_mulhsu(self, ins: "Instruction"):
        INS_NOT_IMPLEMENTED(ins)

    def instruction_mulhu(self, ins: "Instruction"):
        INS_NOT_IMPLEMENTED(ins)

    def instruction_div(self, ins: "Instruction"):
        rd, rs1, rs2 = self.parse_rd_rs_rs(ins)
        self.regs.set(rd, rs1 // rs2)

    def instruction_divu(self, ins: "Instruction"):
        rd, rs1, rs2 = self.parse_rd_rs_rs(ins, signed=False)
        self.regs.set(rd, rs1 // rs2)

    def instruction_rem(self, ins: "Instruction"):
        rd, rs1, rs2 = self.parse_rd_rs_rs(ins)
        self.regs.set(rd, rs1 % rs2)

    def instruction_remu(self, ins: "Instruction"):
        rd, rs1, rs2 = self.parse_rd_rs_rs(ins, signed=False)
        self.regs.set(rd, rs1 % rs2)



"""
RiscEmu (c) 2021 Anton Lydike

SPDX-License-Identifier: MIT
"""

from abc import abstractmethod
from ..colors import *
import typing

if typing.TYPE_CHECKING:
    from . import Instruction


class RiscemuBaseException(BaseException):
    @abstractmethod
    def message(self) -> str:
        raise NotImplemented

    def print_stacktrace(self):
        import traceback

        traceback.print_exception(type(self), self, self.__traceback__)


# Parsing exceptions:


class ParseException(RiscemuBaseException):
    def __init__(self, msg: str, data=None):
        super().__init__(msg, data)
        self.msg = msg
        self.data = data

    def message(self):
        return (
            FMT_PARSE
            + '{}("{}", data={})'.format(self.__class__.__name__, self.msg, self.data)
            + FMT_NONE
        )


def ASSERT_EQ(a1, a2):
    if a1 != a2:
        raise ParseException(
            "ASSERTION_FAILED: Expected elements to be equal!", (a1, a2)
        )


def ASSERT_LEN(a1, size):
    if len(a1) != size:
        raise ParseException(
            "ASSERTION_FAILED: Expected {} to be of length {}".format(a1, size),
            (a1, size),
        )


def ASSERT_NOT_NULL(a1):
    if a1 is None:
        raise ParseException(
            "ASSERTION_FAILED: Expected {} to be non null".format(a1), (a1,)
        )


def ASSERT_NOT_IN(a1, a2):
    if a1 in a2:
        raise ParseException(
            "ASSERTION_FAILED: Expected {} to not be in {}".format(a1, a2), (a1, a2)
        )


def ASSERT_IN(a1, a2):
    if a1 not in a2:
        raise ParseException(
            "ASSERTION_FAILED: Expected {} to not be in {}".format(a1, a2), (a1, a2)
        )


class LinkerException(RiscemuBaseException):
    def __init__(self, msg: str, data):
        self.msg = msg
        self.data = data

    def message(self):
        return (
            FMT_PARSE
            + '{}("{}", data={})'.format(self.__class__.__name__, self.msg, self.data)
            + FMT_NONE
        )


# MMU Exceptions


class MemoryAccessException(RiscemuBaseException):
    def __init__(self, msg: str, addr, size, op):
        super(MemoryAccessException, self).__init__()
        self.msg = msg
        self.addr = addr
        self.size = size
        self.op = op

    def message(self):
        return (
            FMT_MEM
            + "{}(During {} at 0x{:08x} of size {}: {})".format(
                self.__class__.__name__, self.op, self.addr, self.size, self.msg
            )
            + FMT_NONE
        )


class OutOfMemoryException(RiscemuBaseException):
    def __init__(self, action):
        self.action = action

    def message(self):
        return (
            FMT_MEM
            + "{}(Ran out of memory during {})".format(
                self.__class__.__name__, self.action
            )
            + FMT_NONE
        )


class InvalidAllocationException(RiscemuBaseException):
    def __init__(self, msg, name, size, flags):
        self.msg = msg
        self.name = name
        self.size = size
        self.flags = flags

    def message(self):
        return FMT_MEM + "{}[{}](name={}, size={}, flags={})".format(
            self.__class__.__name__, self.msg, self.name, self.size, self.flags
        )


# CPU Exceptions


class UnimplementedInstruction(RiscemuBaseException):
    def __init__(self, ins: "Instruction", context=None):
        self.ins = ins
        self.context = context

    def message(self):
        return (
            FMT_CPU
            + "{}({}{})".format(
                self.__class__.__name__,
                repr(self.ins),
                ", context={}".format(self.context) if self.context is not None else "",
            )
            + FMT_NONE
        )


class InvalidRegisterException(RiscemuBaseException):
    def __init__(self, reg):
        self.reg = reg

    def message(self):
        return (
            FMT_CPU
            + "{}(Invalid register {})".format(self.__class__.__name__, self.reg)
            + FMT_NONE
        )


class InvalidSyscallException(RiscemuBaseException):
    def __init__(self, scall):
        self.scall = scall

    def message(self):
        return (
            FMT_SYSCALL
            + "{}(Invalid syscall: {})".format(self.__class__.__name__, self.scall)
            + FMT_NONE
        )


def INS_NOT_IMPLEMENTED(ins):
    raise UnimplementedInstruction(ins)


class NumberFormatException(RiscemuBaseException):
    def __init__(self, msg):
        super().__init__()
        self.msg = msg

    def message(self):
        return "{}({})".format(self.__class__.__name__, self.msg)


# this exception is not printed and simply signals that an interactive debugging session is
class LaunchDebuggerException(RiscemuBaseException):
    def message(self) -> str:
        return ""



import re
from typing import Union, Tuple
from functools import lru_cache

from . import Instruction, T_RelativeAddress, InstructionContext, Immediate
from ..helpers import parse_numeric_argument

_NUM_LABEL_RE = re.compile(r"[0-9][fb]")
_INT_IMM_RE = re.compile(r"[+-]?([0-9]+|0x[A-Fa-f0-9]+)")


class SimpleInstruction(Instruction):
    def __init__(
        self,
        name: str,
        args: Union[Tuple[()], Tuple[str], Tuple[str, str], Tuple[str, str, str]],
        context: InstructionContext,
        addr: T_RelativeAddress,
    ):
        self.context = context
        self.name = name
        self.args = args
        self._addr = addr

    @property
    def addr(self) -> int:
        return self._addr + self.context.base_address

    @lru_cache(maxsize=None)
    def get_imm(self, num: int) -> Immediate:
        token = self.args[num]

        if _INT_IMM_RE.fullmatch(token):
            value = parse_numeric_argument(token)
            return Immediate(abs_value=value, pcrel_value=value + self.addr)

        # resolve label correctly
        if _NUM_LABEL_RE.fullmatch(token):
            value = self.context.resolve_numerical_label(token, self.addr)
        else:
            value = self.context.resolve_label(token)

        # TODO: make it raise a nice error instead
        if value is None:
            raise NumberFormatException(
                "{} is neither a number now a known symbol!".format(token)
            )
        return Immediate(abs_value=value, pcrel_value=value - self.addr)

    def get_reg(self, num: int) -> str:
        return self.args[num]



from typing import List

from . import (
    MemorySection,
    Instruction,
    InstructionContext,
    MemoryFlags,
    T_RelativeAddress,
)
from .exceptions import MemoryAccessException


class InstructionMemorySection(MemorySection):
    def __init__(
        self,
        instructions: List[Instruction],
        name: str,
        context: InstructionContext,
        owner: str,
        base: int = 0,
    ):
        self.name = name
        self.base = base
        self.context = context
        self.size = len(instructions) * 4
        self.flags = MemoryFlags(True, True)
        self.instructions = instructions
        self.owner = owner

    def read(self, offset: T_RelativeAddress, size: int) -> bytearray:
        raise MemoryAccessException(
            "Cannot read raw bytes from instruction section",
            self.base + offset,
            size,
            "read",
        )

    def write(self, offset: T_RelativeAddress, size: int, data: bytearray):
        raise MemoryAccessException(
            "Cannot write raw bytes to instruction section",
            self.base + offset,
            size,
            "write",
        )

    def read_ins(self, offset: T_RelativeAddress) -> Instruction:
        if offset % 4 != 0:
            raise MemoryAccessException(
                "Unaligned instruction fetch!",
                self.base + offset,
                4,
                "instruction fetch",
            )
        return self.instructions[offset // 4]



from enum import Enum, auto
from typing import List, Optional, Tuple, Union

from riscemu.core.exceptions import ASSERT_LEN, ParseException

from .colors import FMT_NONE, FMT_PARSE
from .core import (
    BinaryDataMemorySection,
    Instruction,
    InstructionContext,
    InstructionMemorySection,
    Int32,
    Program,
    SimpleInstruction,
    T_RelativeAddress,
)
from .helpers import align_addr, get_section_base_name, parse_numeric_argument
from .tokenizer import Token

INSTRUCTION_SECTION_NAMES = (".text", ".init", ".fini")
"""
A tuple containing all section names which contain executable code (instead of data)

The first segment of each segment (first segment of ".text.main" is ".text") is checked
against this list to determine the type of it.
"""


class MemorySectionType(Enum):
    Data = auto()
    Instructions = auto()


class CurrentSection:
    name: str
    data: Union[List[Instruction], bytearray]
    type: MemorySectionType
    base: int

    def __init__(self, name: str, type: MemorySectionType, base: int = 0):
        self.name = name
        self.type = type
        self.base = base
        if self.type == MemorySectionType.Data:
            self.data = bytearray()
        elif self.type == MemorySectionType.Instructions:
            self.data = list()
        else:
            raise ParseException("Unknown section type: {}".format(type))

    def current_address(self) -> T_RelativeAddress:
        if self.type == MemorySectionType.Data:
            return len(self.data) + self.base
        return len(self.data) * 4 + self.base

    def __repr__(self):
        return "{}(name={},data={},type={})".format(
            self.__class__.__name__, self.name, self.data, self.type.name
        )


class ParseContext:
    section: Optional[CurrentSection]
    context: InstructionContext
    program: Program

    def __init__(self, name: str):
        self.program = Program(name)
        self.context = self.program.context
        self.section = None

    def finalize(self) -> Program:
        self._finalize_section()
        return self.program

    def _finalize_section(self):
        if self.section is None:
            return

        if self.section.type == MemorySectionType.Data:
            section = BinaryDataMemorySection(
                self.section.data,
                self.section.name,
                self.context,
                self.program.name,
                self.section.base,
            )
            self.program.add_section(section)
        elif self.section.type == MemorySectionType.Instructions:
            section = InstructionMemorySection(
                self.section.data,
                self.section.name,
                self.context,
                self.program.name,
                self.section.base,
            )
            self.program.add_section(section)

        self.section = None

    def new_section(self, name: str, type: MemorySectionType, alignment: int = 4):
        base = align_addr(self.current_address(), alignment)

        self._finalize_section()
        self.section = CurrentSection(name, type, base)

    def add_label(
        self, name: str, value: int, is_global: bool = False, is_relative: bool = False
    ):
        self.context.labels[name] = value
        if is_global:
            self.program.global_labels.add(name)
        if is_relative:
            self.program.relative_labels.add(name)

    def current_address(self):
        if self.section:
            return self.section.current_address()
        return self.program.base if self.program.base is not None else 0

    def __repr__(self):
        return "{}(\n\tsetion={},\n\tprogram={}\n)".format(
            self.__class__.__name__, self.section, self.program
        )


def ASSERT_IN_SECTION_TYPE(context: ParseContext, type: MemorySectionType):
    if context.section is None:
        raise ParseException(
            "Error, expected to be in {} section, but no section is present...".format(
                type.name
            )
        )
    if context.section.type != type:
        raise ParseException(
            "Error, expected to be in {} section, but currently in {}...".format(
                type.name, context.section
            )
        )


class AssemblerDirectives:
    """
    This class represents a collection of all assembler directives as documented by
    https://github.com/riscv-non-isa/riscv-asm-manual/blob/master/riscv-asm.md#pseudo-ops

    All class methods prefixed with op_ are directly used as assembler directives.
    """

    @classmethod
    def op_align(cls, token: Token, args: Tuple[str], context: ParseContext):
        ASSERT_LEN(args, 1)
        ASSERT_IN_SECTION_TYPE(context, MemorySectionType.Data)
        align_to = parse_numeric_argument(args[0])
        current_mod = context.current_address() % align_to
        if current_mod == 0:
            return
        context.section.data += bytearray(align_to - current_mod)

    @classmethod
    def op_p2align(cls, token: Token, args: Tuple[str], context: ParseContext):
        ASSERT_LEN(args, 1)
        ASSERT_IN_SECTION_TYPE(context, MemorySectionType.Instructions)
        align_to = 2 ** parse_numeric_argument(args[0])

        current_mod = context.current_address() % align_to
        if current_mod == 0:
            return
        num_bytes_fill = align_to - current_mod
        # fill in with nops:
        NOP_SIZE = 4
        for i in range(num_bytes_fill // NOP_SIZE):
            context.section.data.append(
                SimpleInstruction(
                    "nop",
                    (),
                    context.context,
                    context.current_address(),
                )
            )

    @classmethod
    def op_section(cls, token: Token, args: Tuple[str], context: ParseContext):
        ASSERT_LEN(args, 1)
        if get_section_base_name(args[0]) in INSTRUCTION_SECTION_NAMES:
            context.new_section(args[0], MemorySectionType.Instructions)
        else:
            context.new_section(args[0], MemorySectionType.Data)

    @classmethod
    def op_globl(cls, token: Token, args: Tuple[str], context: ParseContext):
        ASSERT_LEN(args, 1)
        context.program.global_labels.add(args[0])

    @classmethod
    def op_global(cls, token: Token, args: Tuple[str], context: ParseContext):
        cls.op_globl(token, args, context)

    @classmethod
    def op_equ(cls, token: Token, args: Tuple[str], context: ParseContext):
        ASSERT_LEN(args, 2)
        name = args[0]
        value = parse_numeric_argument(args[1])
        context.context.labels[name] = value

    @classmethod
    def op_space(cls, token: Token, args: Tuple[str], context: ParseContext):
        ASSERT_LEN(args, 1)
        ASSERT_IN_SECTION_TYPE(context, MemorySectionType.Data)

        size = parse_numeric_argument(args[0])
        cls.add_bytes(size, None, context)

    @classmethod
    def op_zero(cls, token: Token, args: Tuple[str], context: ParseContext):
        ASSERT_LEN(args, 1)
        ASSERT_IN_SECTION_TYPE(context, MemorySectionType.Data)
        size = parse_numeric_argument(args[0])
        cls.add_bytes(size, bytearray(size), context)

    @classmethod
    def add_bytes(
        cls, size: int, content: Union[None, int, bytearray], context: ParseContext
    ):
        ASSERT_IN_SECTION_TYPE(context, MemorySectionType.Data)

        if content is None:
            content = bytearray(size)
        if isinstance(content, int):
            content = Int32(content).to_bytes(size)

        context.section.data += content

    @classmethod
    def add_text(cls, text: str, context: ParseContext, zero_terminate: bool = True):
        # replace '\t' and '\n' escape sequences
        text = text.replace("\\n", "\n").replace("\\t", "\t")

        encoded_bytes = bytearray(text.encode("ascii"))
        if zero_terminate:
            encoded_bytes += bytearray(1)
        cls.add_bytes(len(encoded_bytes), encoded_bytes, context)

    @classmethod
    def handle_instruction(cls, token: Token, args: Tuple[str], context: ParseContext):
        op = token.value[1:]
        if hasattr(cls, "op_" + op):
            getattr(cls, "op_" + op)(token, args, context)
        elif op in ("text", "data", "rodata", "bss", "sbss"):
            cls.op_section(token, (token.value,), context)
        elif op in ("string", "asciiz", "asciz", "ascii"):
            ASSERT_LEN(args, 1)
            cls.add_text(args[0], context, zero_terminate=(op != "ascii"))
        elif op in DATA_OP_SIZES:
            size = DATA_OP_SIZES[op]
            for arg in args:
                cls.add_bytes(size, parse_numeric_argument(arg), context)
        else:
            print(
                FMT_PARSE
                + "Unknown assembler directive: {} {} in {}".format(
                    token, args, context
                )
                + FMT_NONE
            )


DATA_OP_SIZES = {
    "byte": 1,
    "2byte": 2,
    "half": 2,
    "short": 2,
    "4byte": 4,
    "word": 4,
    "long": 4,
    "8byte": 8,
    "dword": 8,
    "quad": 8,
}



import sys

from riscemu.config import RunConfig
from riscemu.core import InstructionMemorySection, SimpleInstruction, Program

if __name__ == "__main__":
    from core.usermode_cpu import UserModeCPU
    from .instructions import InstructionSetDict

    cpu = UserModeCPU(list(InstructionSetDict.values()), RunConfig(verbosity=4))

    program = Program("interactive session", base=0x100)
    context = program.context
    program.add_section(
        InstructionMemorySection(
            [
                SimpleInstruction("ebreak", (), context, 0x100),
                SimpleInstruction("addi", ("a0", "zero", "0"), context, 0x104),
                SimpleInstruction("addi", ("a7", "zero", "93"), context, 0x108),
                SimpleInstruction("scall", (), context, 0x10C),
            ],
            ".text",
            context,
            program.name,
            0x100,
        )
    )

    cpu.load_program(program)

    cpu.setup_stack()

    cpu.launch()

    sys.exit(cpu.exit_code)



from collections import defaultdict

from typing import Dict, Callable, Optional, Union
from . import UInt32, PrivModes, csr_constants
from .traps import InstructionAccessFault

from ..colors import FMT_ERROR, FMT_NONE


def _invalid_setter(addr: int, old_val: UInt32, new_val: UInt32):
    print(FMT_ERROR + f"Cannot write csr 0x{addr:03X}" + FMT_NONE)
    raise InstructionAccessFault(addr)


def _invalid_getter(addr: int, curr_val: UInt32) -> UInt32:
    print(FMT_ERROR + f"Cannot read csr 0x{addr:03X}" + FMT_NONE)
    raise InstructionAccessFault(addr)


class MStatusRegister:
    """
    helper for the mstatus register
    """

    state: UInt32

    def __init__(self):
        self.state = UInt32()

    @property
    def uie(self) -> UInt32:
        return (self.state & (1 << 0)) >> 0

    @uie.setter
    def uie(self, new_val: UInt32):
        new_val = new_val ^ self.uie
        self.state = self.state ^ (new_val << 0)

    @property
    def sie(self) -> UInt32:
        return (self.state & (1 << 1)) >> 1

    @sie.setter
    def sie(self, new_val: UInt32):
        new_val = new_val ^ self.sie
        self.state = self.state ^ (new_val << 1)

    @property
    def mie(self) -> UInt32:
        return (self.state & (1 << 3)) >> 3

    @mie.setter
    def mie(self, new_val: UInt32):
        new_val = new_val ^ self.mie
        self.state = self.state ^ (new_val << 3)

    @property
    def upie(self) -> UInt32:
        return (self.state & (1 << 4)) >> 4

    @upie.setter
    def upie(self, new_val: UInt32):
        new_val = new_val ^ self.upie
        self.state = self.state ^ (new_val << 4)

    @property
    def spie(self) -> UInt32:
        return (self.state & (1 << 5)) >> 5

    @spie.setter
    def spie(self, new_val: UInt32):
        new_val = new_val ^ self.spie
        self.state = self.state ^ (new_val << 5)

    @property
    def mpie(self) -> UInt32:
        return (self.state & (1 << 7)) >> 7

    @mpie.setter
    def mpie(self, new_val: UInt32):
        new_val = new_val ^ self.mpie
        self.state = self.state ^ (new_val << 7)

    @property
    def spp(self) -> UInt32:
        return (self.state & (1 << 8)) >> 8

    @spp.setter
    def spp(self, new_val: UInt32):
        new_val = new_val ^ self.spp
        self.state = self.state ^ (new_val << 8)

    @property
    def mpp(self) -> UInt32:
        # bitwidth = 2
        return (self.state & (0b11 << 11)) >> 11

    @mpp.setter
    def mpp(self, new_val: UInt32):
        new_val = new_val ^ self.mpp
        self.state = self.state ^ (new_val << 11)

    @property
    def fs(self) -> UInt32:
        # bitwidth = 2
        return (self.state & (0b11 << 13)) >> 13

    @fs.setter
    def fs(self, new_val: UInt32):
        new_val = new_val ^ self.fs
        self.state = self.state ^ (new_val << 13)

    @property
    def xs(self) -> UInt32:
        # bitwidth = 2
        return (self.state & (0b11 << 15)) >> 15

    @xs.setter
    def xs(self, new_val: UInt32):
        new_val = new_val ^ self.xs
        self.state = self.state ^ (new_val << 15)

    @property
    def mpriv(self) -> UInt32:
        return (self.state & (1 << 17)) >> 17

    @mpriv.setter
    def mpriv(self, new_val: UInt32):
        new_val = new_val ^ self.mpriv
        self.state = self.state ^ (new_val << 17)

    @property
    def sum(self) -> UInt32:
        return (self.state & (1 << 18)) >> 18

    @sum.setter
    def sum(self, new_val: UInt32):
        new_val = new_val ^ self.sum
        self.state = self.state ^ (new_val << 18)

    @property
    def mxr(self) -> UInt32:
        return (self.state & (1 << 19)) >> 19

    @mxr.setter
    def mxr(self, new_val: UInt32):
        new_val = new_val ^ self.mxr
        self.state = self.state ^ (new_val << 19)

    @property
    def tvm(self) -> UInt32:
        return (self.state & (1 << 20)) >> 20

    @tvm.setter
    def tvm(self, new_val: UInt32):
        new_val = new_val ^ self.tvm
        self.state = self.state ^ (new_val << 20)

    @property
    def tw(self) -> UInt32:
        return (self.state & (1 << 21)) >> 21

    @tw.setter
    def tw(self, new_val: UInt32):
        new_val = new_val ^ self.tw
        self.state = self.state ^ (new_val << 21)

    @property
    def tsr(self) -> UInt32:
        return (self.state & (1 << 22)) >> 22

    @tsr.setter
    def tsr(self, new_val: UInt32):
        new_val = new_val ^ self.tsr
        self.state = self.state ^ (new_val << 22)

    @property
    def sd(self) -> UInt32:
        return (self.state & (1 << 31)) >> 31

    @sd.setter
    def sd(self, new_val: UInt32):
        new_val = new_val ^ self.sd
        self.state = self.state ^ (new_val << 31)


class CSR:
    """
    Represents a processors control and status registers
    """

    state: Dict[int, UInt32]

    setters: Dict[int, Callable[[int, UInt32, UInt32], UInt32]]
    getters: Dict[int, Callable[[int, UInt32], UInt32]]

    mstatus: MStatusRegister

    def __init__(self):
        self.state = defaultdict(UInt32)
        self.mstatus = MStatusRegister()
        # wire mstatus state up to our csr state
        self.state[CSR.name_to_addr("mstatus")] = self.mstatus.state

        self.getters = dict()
        self.setters = dict()

    def get(self, addr: int) -> UInt32:
        if addr in self.getters:
            return self.getters[addr](addr, self.state[addr])
        return self.state[addr]

    def set(self, addr: int, val: UInt32):
        if addr in self.setters:
            self.state[addr] = self.setters[addr](addr, self.state[addr], val)
        else:
            self.state[addr] = val

    def register_callback(
        self,
        addr: int,
        *,
        getter: Optional[Callable[[int, UInt32], UInt32]] = None,
        setter: Optional[Callable[[int, UInt32, UInt32], UInt32]] = None,
    ):
        """
        addr: the CSR address
        getter: a function mapping (addr, old_val_in_store) -> actual value
        setter: a function mapping (addr, old_val_in_store, given_val) -> new_val_in_store

        where the _in_store values denote the values in the CSR store.
        Values may be handled by external stores, if that's the case
        these arguments can be safely ignored.
        """
        if getter is None:
            getter = _invalid_getter
        self.getters[addr] = getter
        if setter is None:
            setter = _invalid_setter
        self.setters[addr] = setter

    @staticmethod
    def assert_can_read(mode: PrivModes, addr: int):
        if (addr >> 8) & 3 > mode.value:
            raise InstructionAccessFault(addr)

    @staticmethod
    def assert_can_write(mode: PrivModes, addr: int):
        if (addr >> 8) & 3 > mode.value or addr >> 10 == 0b11:
            raise InstructionAccessFault(addr)

    @staticmethod
    def name_to_addr(addr: Union[str, int]) -> Optional[int]:
        if isinstance(addr, str):
            if addr not in csr_constants.CSR_NAME_TO_ADDR:
                print("Unknown CSR register {}".format(addr))
                return None
            return csr_constants.CSR_NAME_TO_ADDR[addr]
        return addr



from time import time
from typing import Optional

from . import UInt32


class RTClock:
    """
    Represents a realtime clock. Can be set up to start at a certain time t0, and
    has a certain tickrate.
    """

    tickrate: int
    t0: float

    def __init__(self, tickrate: int, t0: Optional[float] = None):
        if t0 is None:
            self.t0 = time()
        self.tickrate = tickrate

    def get_low32(self, *args) -> UInt32:
        return UInt32(int((time() - self.t0) * self.tickrate))

    def get_hi32(self, *args) -> UInt32:
        return UInt32(int((time() - self.t0) * self.tickrate) >> 32)

    def get_as_int(self):
        return time() * self.tickrate



import json
from collections import defaultdict
from dataclasses import dataclass
from functools import lru_cache
from typing import Tuple, Dict, Set

from riscemu.colors import FMT_NONE, FMT_PARSE
from riscemu.decoder import format_ins, RISCV_REGS, decode
from riscemu.core.traps import (
    InstructionAccessFault,
    InstructionAddressMisalignedTrap,
    LoadAccessFault,
)
from riscemu.core import (
    Instruction,
    InstructionContext,
    T_RelativeAddress,
    MemoryFlags,
    T_AbsoluteAddress,
    BinaryDataMemorySection,
    Immediate,
)


@dataclass(frozen=True)
class ElfInstruction(Instruction):
    name: str
    args: Tuple[int]
    encoded: int

    def get_imm(self, num: int) -> Immediate:
        return Immediate(self.args[num], self.args[num])

    def get_reg(self, num: int) -> str:
        return RISCV_REGS[self.args[num]]

    def __repr__(self) -> str:
        if self.name == "jal" and self.args[0] == 0:
            return "j       {}".format(self.args[1])
        if self.name == "addi" and self.args[2] == 0:
            return "mv      {}, {}".format(self.get_reg(0), self.get_reg(1))
        if self.name == "addi" and self.args[1] == 0:
            return "li      {}, {}".format(self.get_reg(0), self.args[2])
        if self.name == "ret" and len(self.args) == 0:
            return "ret"
        return format_ins(self.encoded, self.name)


class ElfMemorySection(BinaryDataMemorySection):
    def __init__(
        self,
        data: bytearray,
        name: str,
        context: InstructionContext,
        owner: str,
        base: int,
        flags: MemoryFlags,
    ):
        super().__init__(data, name, context, owner, base=base, flags=flags)
        self.read_ins = lru_cache(maxsize=self.size // 4)(self.read_ins)

    def read_ins(self, offset):
        if not self.flags.executable:
            print(
                FMT_PARSE + "Reading instruction from non-executable memory!" + FMT_NONE
            )
            raise InstructionAccessFault(offset + self.base)
        if offset % 4 != 0:
            raise InstructionAddressMisalignedTrap(offset + self.base)
        return ElfInstruction(*decode(self.data[offset : offset + 4]))

    def write(self, offset: T_RelativeAddress, size: int, data: bytearray):
        if self.flags.read_only:
            raise LoadAccessFault(
                "read-only section", offset + self.base, size, "write"
            )
        self.read_ins.cache_clear()
        return super(ElfMemorySection, self).write(offset, size, data)

    @property
    def end(self):
        return self.size + self.base


class MemoryImageDebugInfos:
    VERSION = "1.0.0"
    """
    Schema version
    """

    base: T_AbsoluteAddress = 0
    """
    The base address where the image starts. Defaults to zero.
    """

    sections: Dict[str, Dict[str, Tuple[int, int]]]
    """
    This dictionary maps a program and section to (start address, section length)
    """

    symbols: Dict[str, Dict[str, int]]
    """
    This dictionary maps a program and a symbol to a value
    """

    globals: Dict[str, Set[str]]
    """
    This dictionary contains the list of all global symbols of a given program
    """

    def __init__(
        self,
        sections: Dict[str, Dict[str, Tuple[int, int]]],
        symbols: Dict[str, Dict[str, int]],
        globals: Dict[str, Set[str]],
        base: int = 0,
    ):
        self.sections = sections
        self.symbols = symbols
        self.globals = globals
        for name in globals:
            globals[name] = set(globals[name])
        self.base = base

    def serialize(self) -> str:
        def serialize(obj: any) -> str:
            if isinstance(obj, defaultdict):
                return json.dumps(dict(obj), default=serialize)
            if isinstance(obj, (set, tuple)):
                return json.dumps(list(obj), default=serialize)
            return "<<unserializable {}>>".format(
                getattr(obj, "__qualname__", "{unknown}")
            )

        return json.dumps(
            dict(
                sections=self.sections,
                symbols=self.symbols,
                globals=self.globals,
                base=self.base,
                VERSION=self.VERSION,
            ),
            default=serialize,
        )

    @classmethod
    def load(cls, serialized_str: str) -> "MemoryImageDebugInfos":
        json_obj: dict = json.loads(serialized_str)

        if "VERSION" not in json_obj:
            raise RuntimeError("Unknown MemoryImageDebugInfo version!")

        version: str = json_obj.pop("VERSION")

        # compare major version
        if (
            version != cls.VERSION
            and version.split(".")[0] != cls.VERSION.split(".")[0]
        ):
            raise RuntimeError(
                "Unknown MemoryImageDebugInfo version! This emulator expects version {}, debug info version {}".format(
                    cls.VERSION, version
                )
            )

        return MemoryImageDebugInfos(**json_obj)

    @classmethod
    def builder(cls) -> "MemoryImageDebugInfos":
        return MemoryImageDebugInfos(
            defaultdict(dict), defaultdict(dict), defaultdict(set)
        )



"""
RiscEmu (c) 2021 Anton Lydike

SPDX-License-Identifier: MIT
"""

import re
from dataclasses import dataclass
from enum import Enum, auto
from typing import List, Iterable

from riscemu.decoder import RISCV_REGS
from riscemu.core.exceptions import ParseException

LINE_COMMENT_STARTERS = ("#", ";", "//")
WHITESPACE_PATTERN = re.compile(r"\s+")
MEMORY_ADDRESS_PATTERN = re.compile(
    r"^(0[xX][A-f0-9]+|\d+|0b[0-1]+|[A-z0-9_-]+)\(([A-z]+[0-9]*)\)$"
)
REGISTER_NAMES = RISCV_REGS


class TokenType(Enum):
    COMMA = auto()
    ARGUMENT = auto()
    PSEUDO_OP = auto()
    INSTRUCTION_NAME = auto()
    NEWLINE = auto()
    LABEL = auto()


@dataclass(frozen=True)
class Token:
    type: TokenType
    value: str

    def __str__(self):
        if self.type == TokenType.NEWLINE:
            return "\\n"
        if self.type == TokenType.COMMA:
            return ", "
        return "{}({})".format(self.type.name[0:3], self.value)


NEWLINE = Token(TokenType.NEWLINE, "\n")
COMMA = Token(TokenType.COMMA, ",")


def tokenize(input: Iterable[str]) -> Iterable[Token]:
    for line in input:
        for line_comment_start in LINE_COMMENT_STARTERS:
            if line_comment_start in line:
                line = line[: line.index(line_comment_start)]
        line.strip(" \t\n")
        if not line:
            continue

        parts = list(part for part in split_whitespace_respecting_quotes(line) if part)

        yield from parse_line(parts)
        yield NEWLINE


def parse_line(parts: List[str]) -> Iterable[Token]:
    if len(parts) == 0:
        return ()
    first_token = parts[0]

    if first_token[0] == ".":
        yield Token(TokenType.PSEUDO_OP, first_token)
    elif first_token[-1] == ":":
        yield Token(TokenType.LABEL, first_token)
        yield from parse_line(parts[1:])
        return
    else:
        yield Token(TokenType.INSTRUCTION_NAME, first_token)

    for part in parts[1:]:
        if part == ",":
            yield COMMA
            continue
        yield from parse_arg(part)


def parse_arg(arg: str) -> Iterable[Token]:
    comma = arg[-1] == ","
    arg = arg[:-1] if comma else arg
    mem_match_result = re.match(MEMORY_ADDRESS_PATTERN, arg)
    if mem_match_result:
        register = mem_match_result.group(2).lower()
        immediate = mem_match_result.group(1)
        yield Token(TokenType.ARGUMENT, register)
        yield Token(TokenType.ARGUMENT, immediate)
    else:
        yield Token(TokenType.ARGUMENT, arg)
    if comma:
        yield COMMA


def print_tokens(tokens: Iterable[Token]):
    for token in tokens:
        print(token, end="\n" if token == NEWLINE else "")
    print("", flush=True, end="")


def split_whitespace_respecting_quotes(line: str) -> Iterable[str]:
    quote = ""
    part = ""
    for c in line:
        if c == quote:
            yield part
            part = ""
            quote = ""
            continue

        if quote != "":
            part += c
            continue

        if c in "\"'":
            if part:
                yield part
            quote = c
            part = ""
            continue

        if c in " \t\n":
            if part:
                yield part
            part = ""
            continue

        part += c

    if part:
        yield part



from typing import Dict, Callable, List, Union
from .regs import RISCV_REGS


def op(ins: int):
    return (ins >> 2) & 31


def rd(ins: int):
    return (ins >> 7) & 31


def funct3(ins: int):
    return (ins >> 12) & 7


def rs1(ins: int):
    return (ins >> 15) & 31


def rs2(ins: int):
    return (ins >> 20) & 31


def funct7(ins: int):
    return ins >> 25


def imm110(ins: int):
    return ins >> 20


def imm3112(ins: int):
    return ins >> 12


def imm_i(ins: int):
    return sign_extend(imm110(ins), 12)


def imm_s(ins: int):
    num = (funct7(ins) << 5) + rd(ins)
    return sign_extend(num, 12)


def imm_b(ins: int):
    lower = rd(ins)
    higher = funct7(ins)

    num = (
        (lower & 0b11110)
        + ((higher & 0b0111111) << 5)
        + ((lower & 1) << 11)
        + ((higher >> 6) << 12)
    )
    return sign_extend(num, 13)


def imm_u(ins: int):
    return sign_extend(imm3112(ins), 20)


def imm_j(ins: int):
    return sign_extend(
        (((ins >> 21) & 0b1111111111) << 1)
        + (((ins >> 20) & 1) << 11)
        + (((ins >> 12) & 0b11111111) << 12)
        + (((ins >> 31) & 1) << 20),
        21,
    )


def sign_extend(num, bits):
    sign_mask = 1 << (bits - 1)
    return (num & (sign_mask - 1)) - (num & sign_mask)


def decode_i(ins: int) -> List[int]:
    return [rd(ins), rs1(ins), imm_i(ins)]


def decode_b(ins: int) -> List[int]:
    return [rs1(ins), rs2(ins), imm_b(ins)]


def decode_u(ins: int) -> List[int]:
    return [rd(ins), imm_u(ins)]


def decode_r(ins: int) -> List[int]:
    return [rd(ins), rs1(ins), rs2(ins)]


def decode_s(ins: int) -> List[int]:
    return [rs2(ins), rs1(ins), imm_s(ins)]


def decode_j(ins: int) -> List[int]:
    return [rd(ins), imm_j(ins)]


def decode_i_shamt(ins: int) -> List[int]:
    if funct3(ins) in (1, 5):
        return [rd(ins), rs1(ins), rs2(ins)]
    return decode_i(ins)


def decode_i_unsigned(ins: int) -> List[int]:
    return [rd(ins), rs1(ins), imm110(ins)]


INSTRUCTION_ARGS_DECODER: Dict[int, Callable[[int], List[int]]] = {
    0x00: decode_i,
    0x04: decode_i_shamt,
    0x05: decode_u,
    0x08: decode_s,
    0x0C: decode_r,
    0x0D: decode_u,
    0x18: decode_b,
    0x19: decode_i,
    0x1B: decode_j,
    0x1C: decode_i_unsigned,
    0b1011: decode_r,
}



from abc import ABC, abstractmethod
from typing import List, Type, Callable, Set, Dict, TYPE_CHECKING

from ..config import RunConfig
from ..colors import FMT_NONE, FMT_CPU
from . import (
    T_AbsoluteAddress,
    Instruction,
    Program,
    PrivModes,
    MMU,
    UInt32,
    Registers,
    CSR,
    RTClock,
    csr_constants,
)

if TYPE_CHECKING:
    # from core.mmu import MMU
    from ..instructions import InstructionSet


class CPU(ABC):
    # static cpu configuration
    INS_XLEN: int = 4

    # housekeeping variables
    regs: Registers
    mmu: "MMU"
    pc: T_AbsoluteAddress
    cycle: int
    halted: bool

    # debugging context
    debugger_active: bool

    # instruction information
    instructions: Dict[str, Callable[[Instruction], None]]
    instruction_sets: Set["InstructionSet"]

    # configuration
    conf: RunConfig

    # control and status things:
    hart_id: int
    mode: PrivModes
    csr: CSR
    rtclock: RTClock

    def __init__(
        self,
        mmu: "MMU",
        instruction_sets: List[Type["InstructionSet"]],
        conf: RunConfig,
    ):
        self.mmu = mmu
        self.regs = Registers(conf.unlimited_registers, conf.flen)
        self.conf = conf

        self.instruction_sets = set()
        self.instructions = dict()

        for set_class in instruction_sets:
            ins_set = set_class(self)
            self.instructions.update(ins_set.load())
            self.instruction_sets.add(ins_set)

        self.halted = False
        self.cycle = 0
        self.pc = 0
        self.hart_id = 0
        self.debugger_active = False
        self.csr = CSR()
        self.rtclock = RTClock(conf.rtclock_tickrate)

    def run_instruction(self, ins: Instruction):
        """
        Execute a single instruction

        :param ins: The instruction to execute
        """
        try:
            self.instructions[ins.name](ins)
        except KeyError as ex:
            raise RuntimeError("Unknown instruction: {}".format(ins)) from ex

    def load_program(self, program: Program):
        self.mmu.load_program(program)

    def __repr__(self):
        """
        Returns a representation of the CPU and some of its state.
        """
        return "{}(pc=0x{:08X}, cycle={}, halted={} instructions={})".format(
            self.__class__.__name__,
            self.pc,
            self.cycle,
            self.halted,
            " ".join(s.name for s in self.instruction_sets),
        )

    @abstractmethod
    def step(self, verbose: bool = False):
        pass

    @abstractmethod
    def run(self, verbose: bool = False):
        pass

    def initialize_registers(self):
        # set a0 to the hartid
        self.regs.set("a0", UInt32(self.hart_id))

    def launch(self, verbose: bool = False):
        entrypoint = self.mmu.find_entrypoint()

        if entrypoint is None:
            entrypoint = self.mmu.programs[0].entrypoint

        self.initialize_registers()
        self.setup_csr()

        if self.conf.verbosity > 0:
            print(
                FMT_CPU
                + "[CPU] Started running from {}".format(
                    self.mmu.translate_address(entrypoint)
                )
                + FMT_NONE
            )
        self.pc = entrypoint
        self.run(verbose)

    @property
    def sections(self):
        return self.mmu.sections

    @property
    def programs(self):
        return self.mmu.programs

    def setup_csr(self):
        """
        Set up standard CSR registers, can be hooked into when subclassing to provide
        more information.
        """

        self.csr.set(CSR.name_to_addr("mhartid"), UInt32(self.hart_id))
        self.csr.register_callback(
            CSR.name_to_addr("time"),
            getter=self.rtclock.get_low32,
        )
        self.csr.register_callback(
            CSR.name_to_addr("timeh"),
            getter=self.rtclock.get_hi32,
        )
        self.csr.register_callback(
            CSR.name_to_addr("instret"),
            getter=(lambda csr, _: UInt32(self.cycle)),
        )
        self.csr.register_callback(
            CSR.name_to_addr("instreth"),
            getter=(lambda csr, _: UInt32(self.cycle >> 32)),
        )
        self.csr.register_callback(
            CSR.name_to_addr("cycle"),
            getter=(lambda csr, _: UInt32(self.cycle)),
        )
        self.csr.register_callback(
            CSR.name_to_addr("cycleh"),
            getter=(lambda csr, _: UInt32(self.cycle >> 32)),
        )



"""
RiscEmu (c) 2021 Anton Lydike

SPDX-License-Identifier: MIT
"""

from .instruction_set import InstructionSet, ASSERT_LEN

from ..colors import FMT_DEBUG, FMT_NONE
from ..syscall import Syscall
from ..core import Instruction, Int32, UInt32, UserModeCPU, LaunchDebuggerException


class RV32I(InstructionSet):
    """
    The RV32I instruction set. Some instructions are missing, such as
    fence, fence.i, rdcycle, rdcycleh, rdtime, rdtimeh, rdinstret, rdinstreth
    All atomic read/writes are also not implemented yet

    See https://maxvytech.com/images/RV32I-11-2018.pdf for a more detailed overview
    """

    def instruction_lb(self, ins: "Instruction"):
        rd, addr = self.parse_mem_ins(ins)
        self.regs.set(rd, UInt32.sign_extend(self.mmu.read(addr.unsigned_value, 1), 8))

    def instruction_lh(self, ins: "Instruction"):
        rd, addr = self.parse_mem_ins(ins)
        self.regs.set(rd, UInt32.sign_extend(self.mmu.read(addr.unsigned_value, 2), 16))

    def instruction_lw(self, ins: "Instruction"):
        rd, addr = self.parse_mem_ins(ins)
        self.regs.set(rd, UInt32(self.mmu.read(addr.unsigned_value, 4)))

    def instruction_lbu(self, ins: "Instruction"):
        rd, addr = self.parse_mem_ins(ins)
        self.regs.set(rd, UInt32(self.mmu.read(addr.unsigned_value, 1)))

    def instruction_lhu(self, ins: "Instruction"):
        rd, addr = self.parse_mem_ins(ins)
        self.regs.set(rd, UInt32(self.mmu.read(addr.unsigned_value, 2)))

    def instruction_sb(self, ins: "Instruction"):
        rd, addr = self.parse_mem_ins(ins)
        self.mmu.write(addr.unsigned_value, 1, self.regs.get(rd).to_bytes(1))

    def instruction_sh(self, ins: "Instruction"):
        rd, addr = self.parse_mem_ins(ins)
        self.mmu.write(addr.unsigned_value, 2, self.regs.get(rd).to_bytes(2))

    def instruction_sw(self, ins: "Instruction"):
        rd, addr = self.parse_mem_ins(ins)
        self.mmu.write(addr.unsigned_value, 4, self.regs.get(rd).to_bytes(4))

    def instruction_sll(self, ins: "Instruction"):
        ASSERT_LEN(ins.args, 3)
        dst = ins.get_reg(0)
        src1 = ins.get_reg(1)
        src2 = ins.get_reg(2)
        self.regs.set(dst, self.regs.get(src1) << (self.regs.get(src2) & 0b11111))

    def instruction_slli(self, ins: "Instruction"):
        ASSERT_LEN(ins.args, 3)
        dst = ins.get_reg(0)
        src1 = ins.get_reg(1)
        imm = ins.get_imm(2)
        self.regs.set(dst, self.regs.get(src1) << (imm.abs_value & 0b11111))

    def instruction_srl(self, ins: "Instruction"):
        ASSERT_LEN(ins.args, 3)
        dst = ins.get_reg(0)
        src1 = ins.get_reg(1)
        src2 = ins.get_reg(2)
        self.regs.set(
            dst, self.regs.get(src1).shift_right_logical(self.regs.get(src2) & 0b11111)
        )

    def instruction_srli(self, ins: "Instruction"):
        ASSERT_LEN(ins.args, 3)
        dst = ins.get_reg(0)
        src1 = ins.get_reg(1)
        imm = ins.get_imm(2)
        self.regs.set(
            dst, self.regs.get(src1).shift_right_logical(imm.abs_value & 0b11111)
        )

    def instruction_sra(self, ins: "Instruction"):
        ASSERT_LEN(ins.args, 3)
        dst = ins.get_reg(0)
        src1 = ins.get_reg(1)
        src2 = ins.get_reg(2)
        self.regs.set(dst, self.regs.get(src1) >> (self.regs.get(src2) & 0b11111))

    def instruction_srai(self, ins: "Instruction"):
        ASSERT_LEN(ins.args, 3)
        dst = ins.get_reg(0)
        src1 = ins.get_reg(1)
        imm = ins.get_imm(2)
        self.regs.set(dst, self.regs.get(src1) >> (imm.abs_value & 0b11111))

    def instruction_add(self, ins: "Instruction"):
        dst, rs1, rs2 = self.parse_rd_rs_rs(ins)
        self.regs.set(dst, rs1 + rs2)

    def instruction_addi(self, ins: "Instruction"):
        dst, rs1, imm = self.parse_rd_rs_imm(ins)
        self.regs.set(dst, rs1 + imm.abs_value)

    def instruction_sub(self, ins: "Instruction"):
        dst, rs1, rs2 = self.parse_rd_rs_rs(ins)
        self.regs.set(dst, rs1 - rs2)

    def instruction_lui(self, ins: "Instruction"):
        ASSERT_LEN(ins.args, 2)
        reg = ins.get_reg(0)
        self.regs.set(reg, ins.get_imm(1).abs_value << 12)

    def instruction_auipc(self, ins: "Instruction"):
        ASSERT_LEN(ins.args, 2)
        reg = ins.get_reg(0)
        imm = ins.get_imm(1).abs_value << 12
        self.regs.set(reg, imm + self.cpu.pc)

    def instruction_xor(self, ins: "Instruction"):
        rd, rs1, rs2 = self.parse_rd_rs_rs(ins)
        self.regs.set(rd, rs1 ^ rs2)

    def instruction_xori(self, ins: "Instruction"):
        rd, rs1, imm = self.parse_rd_rs_imm(ins)
        self.regs.set(rd, rs1 ^ imm.abs_value)

    def instruction_or(self, ins: "Instruction"):
        rd, rs1, rs2 = self.parse_rd_rs_rs(ins)
        self.regs.set(rd, rs1 | rs2)

    def instruction_ori(self, ins: "Instruction"):
        rd, rs1, imm = self.parse_rd_rs_imm(ins)
        self.regs.set(rd, rs1 | imm.abs_value)

    def instruction_and(self, ins: "Instruction"):
        rd, rs1, rs2 = self.parse_rd_rs_rs(ins)
        self.regs.set(rd, rs1 & rs2)

    def instruction_andi(self, ins: "Instruction"):
        rd, rs1, imm = self.parse_rd_rs_imm(ins)
        self.regs.set(rd, rs1 & imm.abs_value)

    def instruction_slt(self, ins: "Instruction"):
        rd, rs1, rs2 = self.parse_rd_rs_rs(ins)
        self.regs.set(rd, UInt32(rs1 < rs2))

    def instruction_slti(self, ins: "Instruction"):
        rd, rs1, imm = self.parse_rd_rs_imm(ins)
        self.regs.set(rd, UInt32(rs1 < imm))

    def instruction_sltu(self, ins: "Instruction"):
        dst, rs1, rs2 = self.parse_rd_rs_rs(ins)
        self.regs.set(dst, UInt32(rs1.unsigned_value < rs2.unsigned_value))

    def instruction_sltiu(self, ins: "Instruction"):
        dst, rs1, imm = self.parse_rd_rs_imm(ins)
        self.regs.set(dst, UInt32(rs1.unsigned_value < imm.abs_value.unsigned_value))

    def instruction_beq(self, ins: "Instruction"):
        rs1, rs2, dst = self.parse_rs_rs_imm(ins)
        if rs1 == rs2:
            self.cpu.pc += dst.pcrel_value.value - 4

    def instruction_bne(self, ins: "Instruction"):
        rs1, rs2, dst = self.parse_rs_rs_imm(ins)
        if rs1 != rs2:
            self.cpu.pc += dst.pcrel_value.value - 4

    def instruction_blt(self, ins: "Instruction"):
        rs1, rs2, dst = self.parse_rs_rs_imm(ins)
        if rs1 < rs2:
            self.cpu.pc += dst.pcrel_value.value - 4

    def instruction_bge(self, ins: "Instruction"):
        rs1, rs2, dst = self.parse_rs_rs_imm(ins)
        if rs1 >= rs2:
            self.cpu.pc += dst.pcrel_value.value - 4

    def instruction_bltu(self, ins: "Instruction"):
        rs1, rs2, dst = self.parse_rs_rs_imm(ins)
        if rs1.unsigned_value < rs2.unsigned_value:
            self.cpu.pc += dst.pcrel_value.value - 4

    def instruction_bgeu(self, ins: "Instruction"):
        rs1, rs2, dst = self.parse_rs_rs_imm(ins)
        if rs1.unsigned_value >= rs2.unsigned_value:
            self.cpu.pc += dst.pcrel_value.value - 4

    def instruction_j(self, ins: "Instruction"):
        ASSERT_LEN(ins.args, 1)
        addr = ins.get_imm(0)
        self.cpu.pc += addr.pcrel_value.value - 4

    def instruction_jal(self, ins: "Instruction"):
        reg = "ra"  # default register is ra
        if len(ins.args) == 1:
            addr = ins.get_imm(0)
        else:
            ASSERT_LEN(ins.args, 2)
            reg = ins.get_reg(0)
            addr = ins.get_imm(1)
        self.regs.set(reg, UInt32(self.cpu.pc))
        self.cpu.pc += addr.pcrel_value.value - 4

    def instruction_jalr(self, ins: "Instruction"):
        ASSERT_LEN(ins.args, 3)
        reg = ins.get_reg(0)
        base = ins.get_reg(1)
        addr = ins.get_imm(2).abs_value.value
        self.regs.set(reg, Int32(self.cpu.pc))
        self.cpu.pc = self.regs.get(base).unsigned_value + addr

    def instruction_ret(self, ins: "Instruction"):
        ASSERT_LEN(ins.args, 0)
        self.cpu.pc = self.regs.get("ra").unsigned_value

    def instruction_ecall(self, ins: "Instruction"):
        self.instruction_scall(ins)

    def instruction_ebreak(self, ins: "Instruction"):
        self.instruction_sbreak(ins)

    def instruction_scall(self, ins: "Instruction"):
        ASSERT_LEN(ins.args, 0)

        if not isinstance(self.cpu, UserModeCPU):
            # FIXME: add exception for syscall not supported or something
            raise

        syscall = Syscall(self.regs.get("a7"), self.cpu)
        self.cpu.syscall_int.handle_syscall(syscall)

    def instruction_sbreak(self, ins: "Instruction"):
        ASSERT_LEN(ins.args, 0)
        if self.cpu.conf.debug_instruction:
            print(
                FMT_DEBUG
                + "Debug instruction encountered at 0x{:08X}".format(self.cpu.pc - 1)
                + FMT_NONE
            )
            raise LaunchDebuggerException()

    def instruction_nop(self, ins: "Instruction"):
        ASSERT_LEN(ins.args, 0)
        pass

    def instruction_li(self, ins: "Instruction"):
        ASSERT_LEN(ins.args, 2)
        reg = ins.get_reg(0)
        immediate = ins.get_imm(1).abs_value
        self.regs.set(reg, Int32(immediate))

    def instruction_la(self, ins: "Instruction"):
        ASSERT_LEN(ins.args, 2)
        reg = ins.get_reg(0)
        immediate = ins.get_imm(1).abs_value
        self.regs.set(reg, Int32(immediate))

    def instruction_mv(self, ins: "Instruction"):
        ASSERT_LEN(ins.args, 2)
        rd, rs = ins.get_reg(0), ins.get_reg(1)
        self.regs.set(rd, self.regs.get(rs))



from collections import defaultdict
from typing import Dict, List, Optional

from .exceptions import ParseException
from ..core import T_AbsoluteAddress, T_RelativeAddress, NUMBER_SYMBOL_PATTERN


class InstructionContext:
    base_address: T_AbsoluteAddress
    """
    The address where the instruction block is placed
    """

    labels: Dict[str, T_RelativeAddress]
    """
    This dictionary maps all labels to their relative position of the instruction block
    """

    numbered_labels: Dict[str, List[T_RelativeAddress]]
    """
    This dictionary maps numbered labels (which can occur multiple times) to a list of (block-relative) addresses where
    the label was placed
    """

    global_symbol_dict: Dict[str, T_AbsoluteAddress]
    """
    A reference to the MMU's global symbol dictionary for access to global symbols
    """

    def __init__(self):
        self.labels = dict()
        self.numbered_labels = defaultdict(list)
        self.base_address = 0
        self.global_symbol_dict = dict()

    def resolve_numerical_label(
        self, symbol: str, address_at: int
    ) -> Optional[T_AbsoluteAddress]:
        direction = symbol[-1]
        values = self.numbered_labels.get(symbol[:-1], [])
        if direction == "b":
            return max(
                (addr + self.base_address for addr in values if addr < address_at),
                default=None,
            )
        else:
            return min(
                (addr + self.base_address for addr in values if addr > address_at),
                default=None,
            )

    def resolve_label(self, symbol: str) -> Optional[T_AbsoluteAddress]:
        # if it's not a local symbol, try the globals
        if symbol not in self.labels:
            return self.global_symbol_dict.get(symbol, None)
        # otherwise return the local symbol
        return self.labels.get(symbol, None)



"""
RiscEmu (c) 2023 Anton Lydike

SPDX-License-Identifier: MIT

This file contains copious amounts of docstrings that were all taken
from https://msyksphinz-self.github.io/riscv-isadoc/html/rvfd.html
(all the docstrings on the instruction methods documenting the opcodes
and their function)
"""
from typing import Tuple

from .instruction_set import InstructionSet, Instruction
from .float_base import FloatArithBase
from riscemu.core import INS_NOT_IMPLEMENTED, Float32, Int32, UInt32, Float64


class RV32D(FloatArithBase[Float64]):
    flen = 64
    _float_cls = Float64



"""
RiscEmu (c) 2021 Anton Lydike

SPDX-License-Identifier: MIT

The priv Module holds everything necessary for emulating privileged risc-v assembly


Running priv is only preferable to the normal userspace emulator, if you actually want to emulate the whole system.

Syscalls will have to be intercepted by your assembly code.


The PrivCPU Implements the Risc-V M/U Model, meaning there is machine mode and user mode. No PMP or paging is available.
"""



"""
RiscEmu (c) 2021 Anton Lydike

SPDX-License-Identifier: MIT
"""

import sys
from dataclasses import dataclass
from math import log2, ceil
from typing import Dict, IO, Union

from .core import (
    BinaryDataMemorySection,
    MemoryFlags,
    Int32,
    CPU,
    InvalidSyscallException,
)
from .colors import FMT_SYSCALL, FMT_NONE

SYSCALLS = {
    63: "read",
    64: "write",
    93: "exit",
    192: "mmap2",
    1024: "open",
    1025: "close",
}
"""
This dict contains a mapping for all available syscalls (code->name)

If you wish to add a syscall to the built-in system, you can extend this
dictionary and implement a method with the same name on the SyscallInterface
class.
"""

ADDITIONAL_SYMBOLS = {
    "MAP_PRIVATE": 1 << 0,
    "MAP_SHARED": 1 << 1,
    "MAP_ANON": 1 << 2,
    "MAP_ANONYMOUS": 1 << 2,
    "PROT_READ": 1 << 0,
    "PROT_WRITE": 1 << 1,
}
"""
A set of additional symbols that are used by various syscalls.
"""

OPEN_MODES = {
    0: "rb",
    1: "wb",
    2: "r+b",
    3: "x",
    4: "ab",
}
"""All available file open modes"""


@dataclass(frozen=True)
class Syscall:
    """
    Represents a syscall
    """

    id: int
    """The syscall number (e.g. 64 - write)"""
    cpu: CPU
    """The CPU object that created the syscall"""

    @property
    def name(self):
        return SYSCALLS.get(self.id, "unknown")

    def __repr__(self):
        return "Syscall(id={}, name={})".format(self.id, self.name)

    def ret(self, code: Union[int, Int32]):
        self.cpu.regs.set("a0", Int32(code))


def get_syscall_symbols():
    """
    Generate global syscall symbols (such as SCALL_READ, SCALL_EXIT etc)

    :return: dictionary of all syscall symbols (SCALL_<name> -> id)
    """
    items: Dict[str, int] = {
        ("SCALL_" + name.upper()): num for num, name in SYSCALLS.items()
    }

    items.update(ADDITIONAL_SYMBOLS)

    return items


class SyscallInterface:
    """
    Handles syscalls
    """

    open_files: Dict[int, IO]
    next_open_handle: int

    def handle_syscall(self, scall: Syscall):
        self.next_open_handle = 3
        self.open_files = {0: sys.stdin, 1: sys.stdout, 2: sys.stderr}

        if getattr(self, scall.name):
            getattr(self, scall.name)(scall)
        else:
            raise InvalidSyscallException(scall)

    def read(self, scall: Syscall):
        """
        read syscall (63): read from file no a0, into addr a1, at most a2 bytes
        on return a0 will be the number of read bytes or -1 if an error occurred
        """
        fileno = scall.cpu.regs.get("a0").unsigned_value
        addr = scall.cpu.regs.get("a1").unsigned_value
        size = scall.cpu.regs.get("a2").unsigned_value
        if fileno not in self.open_files:
            scall.ret(-1)
            return

        chars = self.open_files[fileno].readline(size)
        try:
            data = bytearray(chars, "ascii")
            scall.cpu.mmu.write(addr, len(data), data)
            return scall.ret(len(data))

        except UnicodeEncodeError:
            print(
                FMT_SYSCALL
                + '[Syscall] read: UnicodeError - invalid input "{}"'.format(chars)
                + FMT_NONE
            )
            return scall.ret(-1)

    def write(self, scall: Syscall):
        """
        write syscall (64): write a2 bytes from addr a1 into fileno a0
        on return a0 will hold the number of bytes written or -1 if an error occurred
        """
        fileno = scall.cpu.regs.get("a0").unsigned_value
        addr = scall.cpu.regs.get("a1").unsigned_value
        size = scall.cpu.regs.get("a2").unsigned_value
        if fileno not in self.open_files:
            return scall.ret(-1)

        data = scall.cpu.mmu.read(addr, size)

        if not isinstance(data, bytearray):
            print(
                FMT_SYSCALL
                + "[Syscall] write: writing from .text region not supported."
                + FMT_NONE
            )
            return scall.ret(-1)

        self.open_files[fileno].write(data.decode("ascii"))
        return scall.ret(size)

    def open(self, scall: Syscall):
        """
        open syscall (1024): read path of a2 bytes from addr a1, in mode a0
        returns the file no in a0

        modes:
            - 0: read
            - 1: write (truncate)
            - 2: read/write (no truncate)
            - 3: only create
            - 4: append

        Requires running with flag scall-fs
        """
        # FIXME: this should be toggleable in a global setting or something
        if True:
            print(
                FMT_SYSCALL
                + "[Syscall] open: opening files not supported without scall-fs flag!"
                + FMT_NONE
            )
            return scall.ret(-1)

        mode = scall.cpu.regs.get("a0").unsigned_value
        addr = scall.cpu.regs.get("a1").unsigned_value
        size = scall.cpu.regs.get("a2").unsigned_value

        mode_st = OPEN_MODES.get(
            mode,
        )
        if mode_st == -1:
            print(
                FMT_SYSCALL
                + "[Syscall] open: unknown opening mode {}!".format(mode)
                + FMT_NONE
            )
            return scall.ret(-1)

        path = scall.cpu.mmu.read(addr, size).decode("ascii")

        fileno = self.next_open_handle
        self.next_open_handle += 1

        try:
            self.open_files[fileno] = open(path, mode_st)
        except OSError as err:
            print(
                FMT_SYSCALL
                + "[Syscall] open: encountered error during {}!".format(err.strerror)
                + FMT_NONE
            )
            return scall.ret(-1)

        print(
            FMT_SYSCALL
            + "[Syscall] open: opened fd {} to {}!".format(fileno, path)
            + FMT_NONE
        )
        return scall.ret(fileno)

    def close(self, scall: Syscall):
        """
        close syscall (1025): closes file no a0

        return -1 if an error was encountered, otherwise returns 0
        """
        fileno = scall.cpu.regs.get("a0").unsigned_value
        if fileno not in self.open_files:
            print(
                FMT_SYSCALL
                + "[Syscall] close: unknown fileno {}!".format(fileno)
                + FMT_NONE
            )
            return scall.ret(-1)

        self.open_files[fileno].close()
        print(FMT_SYSCALL + "[Syscall] close: closed fd {}!".format(fileno) + FMT_NONE)
        del self.open_files[fileno]
        return scall.ret(0)

    def exit(self, scall: Syscall):
        """
        Exit syscall. Exits the system with status code a0
        """
        scall.cpu.halted = True
        scall.cpu.exit_code = scall.cpu.regs.get("a0").signed().value

    def mmap2(self, scall: Syscall):
        """
        mmap2 syscall:

        void *mmap(void *addr, size_t length, int prot, int flags,
                  int fd, off_t offset);

        Only supported modes:
        addr  = <any>
        prot  = either PROT_READ or PROT_READ | PROT_WRITE
        flags = MAP_PRIVATE | MAP_ANONYMOUS
        fd    = <ignored>
        off_t = <ignored>
        """
        addr = scall.cpu.regs.get("a0").unsigned_value
        size = scall.cpu.regs.get("a1").unsigned_value
        prot = scall.cpu.regs.get("a2").unsigned_value
        flags = scall.cpu.regs.get("a3").unsigned_value

        # error out if prot is not 1 or 3:
        # 1 = PROT_READ
        # 3 = PROT_READ | PROT_WRITE
        if prot != 1 and prot != 3:
            return scall.ret(-1)

        # round size up to multiple of 4096
        size = 4096 * ceil(size / 4096)
        section = BinaryDataMemorySection(
            bytearray(size),
            ".data.runtime-allocated",
            None,
            "system",
            base=addr,
            flags=MemoryFlags(read_only=prot != 3, executable=False),
        )

        # try to insert section
        if scall.cpu.mmu.load_section(section, addr != 0):
            return scall.ret(section.base)
        # if that failed, and we tried to force an address,
        # try again at any address
        elif addr != 0:
            section.base = 0
            if scall.cpu.mmu.load_section(section):
                return scall.ret(section.base)
        # if that didn't work, return error
        return scall.ret(-1)

    def __repr__(self):
        return "{}(\n\tfiles={}\n)".format(self.__class__.__name__, self.open_files)



"""
RiscEmu (c) 2021 Anton Lydike

SPDX-License-Identifier: MIT
"""

# Colors

FMT_RED = "\033[31m"
FMT_ORANGE = "\033[33m"
FMT_GRAY = "\033[37m"
FMT_CYAN = "\033[36m"
FMT_GREEN = "\033[32m"
FMT_MAGENTA = "\033[35m"
FMT_BLUE = "\033[34m"
FMT_YELLOW = "\033[93m"

FMT_BOLD = "\033[1m"
FMT_NONE = "\033[0m"
FMT_UNDERLINE = "\033[4m"

FMT_ERROR = FMT_RED + FMT_BOLD

FMT_MEM = FMT_CYAN + FMT_BOLD
FMT_PARSE = FMT_CYAN + FMT_BOLD
FMT_CPU = FMT_BLUE + FMT_BOLD
FMT_SYSCALL = FMT_YELLOW + FMT_BOLD
FMT_DEBUG = FMT_MAGENTA + FMT_BOLD
FMT_CSR = FMT_ORANGE + FMT_BOLD



from typing import Dict, Any
import re

# define some base type aliases so we can keep track of absolute and relative addresses
T_RelativeAddress = int
T_AbsoluteAddress = int

# parser options are just dictionaries with arbitrary values
T_ParserOpts = Dict[str, Any]

NUMBER_SYMBOL_PATTERN = re.compile(r"^\d+[fb]$")

# exceptions
from .exceptions import (
    ParseException,
    NumberFormatException,
    MemoryAccessException,
    OutOfMemoryException,
    LinkerException,
    LaunchDebuggerException,
    RiscemuBaseException,
    InvalidRegisterException,
    InvalidAllocationException,
    InvalidSyscallException,
    UnimplementedInstruction,
    INS_NOT_IMPLEMENTED,
)

# base classes
from .flags import MemoryFlags
from .int32 import UInt32, Int32
from .float import BaseFloat, Float32, Float64
from .rtclock import RTClock
from .instruction import Instruction, Immediate, InstructionWithEncoding
from .instruction_context import InstructionContext
from .memory_section import MemorySection
from .program import Program
from .program_loader import ProgramLoader
from .privmodes import PrivModes
from .mmu import MMU
from .csr import CSR
from .registers import Registers
from .cpu import CPU
from .simple_instruction import SimpleInstruction
from .instruction_memory_section import InstructionMemorySection
from .binary_data_memory_section import BinaryDataMemorySection
from .usermode_cpu import UserModeCPU

__all__ = [
    "T_RelativeAddress",
    "T_AbsoluteAddress",
    "T_ParserOpts",
    "NUMBER_SYMBOL_PATTERN",
    "ParseException",
    "NumberFormatException",
    "MemoryAccessException",
    "OutOfMemoryException",
    "LinkerException",
    "LaunchDebuggerException",
    "RiscemuBaseException",
    "InvalidRegisterException",
    "InvalidAllocationException",
    "InvalidSyscallException",
    "UnimplementedInstruction",
    "INS_NOT_IMPLEMENTED",
    "MemoryFlags",
    "UInt32",
    "Int32",
    "BaseFloat",
    "Float32",
    "Float64",
    "RTClock",
    "Instruction",
    "Immediate",
    "InstructionWithEncoding",
    "InstructionContext",
    "MemorySection",
    "Program",
    "ProgramLoader",
    "PrivModes",
    "MMU",
    "CSR",
    "Registers",
    "CPU",
    "SimpleInstruction",
    "InstructionMemorySection",
    "BinaryDataMemorySection",
    "UserModeCPU",
]



from .instruction_set import InstructionSet, Instruction
from riscemu.core.exceptions import INS_NOT_IMPLEMENTED
from ..core import Int32, UInt32


class RV32A(InstructionSet):
    """
    The RV32A instruction set. Currently, load-reserved and store conditionally are not supported
    due to limitations in the way the MMU is implemented. Maybe a later implementation will add support
    for this?
    """

    def instruction_lr_w(self, ins: "Instruction"):
        INS_NOT_IMPLEMENTED(ins)

    def instruction_sc_w(self, ins: "Instruction"):
        INS_NOT_IMPLEMENTED(ins)

    def instruction_amoswap_w(self, ins: "Instruction"):
        dest, addr, val = self.parse_rd_rs_rs(ins)
        if dest == "zero":
            self.mmu.write(addr.unsigned_value, val.to_bytes())
        else:
            old = Int32(self.mmu.read(addr.unsigned_value, 4))
            self.mmu.write(addr.unsigned_value, val.to_bytes())
            self.regs.set(dest, old)

    def instruction_amoadd_w(self, ins: "Instruction"):
        dest, addr, val = self.parse_rd_rs_rs(ins)
        old = Int32(self.mmu.read(addr.unsigned_value, 4))
        self.mmu.write(addr.unsigned_value, (old + val).to_bytes(4))
        self.regs.set(dest, old)

    def instruction_amoand_w(self, ins: "Instruction"):
        dest, addr, val = self.parse_rd_rs_rs(ins)
        old = Int32(self.mmu.read(addr.unsigned_value, 4))
        self.mmu.write(addr.unsigned_value, (old & val).to_bytes(4))
        self.regs.set(dest, old)

    def instruction_amoor_w(self, ins: "Instruction"):
        dest, addr, val = self.parse_rd_rs_rs(ins)
        old = Int32(self.mmu.read(addr.unsigned_value, 4))
        self.mmu.write(addr.unsigned_value, (old | val).to_bytes(4))
        self.regs.set(dest, old)

    def instruction_amoxor_w(self, ins: "Instruction"):
        dest, addr, val = self.parse_rd_rs_rs(ins)
        old = Int32(self.mmu.read(addr.unsigned_value, 4))
        self.mmu.write(addr.unsigned_value, (old ^ val).to_bytes(4))
        self.regs.set(dest, old)

    def instruction_amomax_w(self, ins: "Instruction"):
        dest, addr, val = self.parse_rd_rs_rs(ins)
        old = Int32(self.mmu.read(addr.unsigned_value, 4))
        self.mmu.write(addr.unsigned_value, max(old, val).to_bytes(4))
        self.regs.set(dest, old)

    def instruction_amomaxu_w(self, ins: "Instruction"):
        val: UInt32
        dest, addr, val = self.parse_rd_rs_rs(ins)
        old = UInt32(self.mmu.read(addr.unsigned_value, 4))

        self.mmu.write(addr.unsigned_value, max(old, val.unsigned()).to_bytes())
        self.regs.set(dest, old)

    def instruction_amomin_w(self, ins: "Instruction"):
        dest, addr, val = self.parse_rd_rs_rs(ins)
        old = Int32(self.mmu.read(addr.unsigned_value, 4))
        self.mmu.write(addr.unsigned_value, min(old, val).to_bytes(4))
        self.regs.set(dest, old)

    def instruction_amominu_w(self, ins: "Instruction"):
        val: UInt32
        dest, addr, val = self.parse_rd_rs_rs(ins)
        old = UInt32(self.mmu.read(addr.unsigned_value, 4))

        self.mmu.write(addr.unsigned_value, min(old, val.unsigned()).to_bytes(4))
        self.regs.set(dest, old)



from dataclasses import dataclass


@dataclass(frozen=True)
class MemoryFlags:
    read_only: bool
    executable: bool

    def __repr__(self):
        return "r{}{}".format(
            "-" if self.read_only else "w", "x" if self.executable else "-"
        )



"""
RiscEmu (c) 2021 Anton Lydike

SPDX-License-Identifier: MIT
"""
import os.path

from .core import SimpleInstruction
from .helpers import *

if typing.TYPE_CHECKING:
    pass

HIST_FILE = os.path.join(os.path.expanduser("~"), ".riscemu_history")


def launch_debug_session(cpu: "CPU", prompt=""):
    if cpu.debugger_active:
        return
    import code
    import readline
    import rlcompleter

    # set the active debug flag
    cpu.debugger_active = True

    # setup some aliases:
    registers = cpu.regs
    regs = cpu.regs
    memory = cpu.mmu
    mem = cpu.mmu
    mmu = cpu.mmu

    # setup helper functions:
    def dump(what, *args, **kwargs):
        if what == regs:
            regs.dump(*args, **kwargs)
        else:
            mmu.dump(what, *args, **kwargs)

    def dump_stack(*args, **kwargs):
        mmu.dump(regs.get("sp"), *args, **kwargs)

    def ins():
        print("Current instruction at 0x{:08X}:".format(cpu.pc))
        return mmu.read_ins(cpu.pc)

    def run_ins(name, *args: str):
        if len(args) > 3:
            print("Invalid arg count!")
            return
        context = mmu.context_for(cpu.pc)

        ins = SimpleInstruction(name, tuple(args), context, cpu.pc)
        print(FMT_DEBUG + "Running instruction {}".format(ins) + FMT_NONE)
        cpu.run_instruction(ins)

    def cont(verbose=False):
        try:
            cpu.run(verbose)
        except LaunchDebuggerException:
            print(FMT_DEBUG + "Returning to debugger...")
            return

    def step():
        try:
            cpu.step(verbose=True)
        except LaunchDebuggerException:
            return

    # collect all variables
    sess_vars = globals()
    sess_vars.update(locals())

    # add tab completion
    readline.set_completer(rlcompleter.Completer(sess_vars).complete)
    readline.parse_and_bind("tab: complete")
    if os.path.exists(HIST_FILE):
        readline.read_history_file(HIST_FILE)

    relaunch_debugger = False

    try:
        code.InteractiveConsole(sess_vars).interact(
            banner=FMT_DEBUG + prompt + FMT_NONE,
            exitmsg="Exiting debugger",
        )
    finally:
        cpu.debugger_active = False
        readline.write_history_file(HIST_FILE)



import argparse
import sys
from dataclasses import dataclass
from io import IOBase, RawIOBase, TextIOBase
from typing import Type, Dict, List, Optional, Union
import importlib_resources

from . import __version__, __copyright__
from .core import CPU, ProgramLoader, Program, UserModeCPU
from .instructions import InstructionSet, InstructionSetDict
from .config import RunConfig
from .helpers import FMT_GRAY, FMT_NONE
from .parser import AssemblyFileLoader
from .instructions.float_base import FloatArithBase


@dataclass
class RiscemuSource:
    name: str
    stream: Union[TextIOBase, RawIOBase]

    def get_score_for(self, loader: ProgramLoader) -> float:
        if loader.is_binary:
            if isinstance(self.stream, TextIOBase):
                return 0
        elif isinstance(self.stream, RawIOBase):
            return 0
        return loader.can_parse(self.name)


class RiscemuMain:
    """
    This represents the riscemu API exposed to other programs for better
    interoperability.
    """

    available_ins_sets: Dict[str, Type[InstructionSet]]
    available_file_loaders: List[Type[ProgramLoader]]

    cfg: Optional[RunConfig]
    cpu: Optional[CPU]

    input_files: List[Union[str, RiscemuSource]]
    selected_ins_sets: List[Type[InstructionSet]]

    def __init__(self, cfg: Optional[RunConfig] = None):
        self.available_ins_sets = dict()
        self.selected_ins_sets = []
        self.available_file_loaders = []
        self.cfg: Optional[RunConfig] = cfg
        self.cpu: Optional[CPU] = None
        self.input_files = []
        self.selected_ins_sets = []

    def instantiate_cpu(self):
        self.cpu = UserModeCPU(self.selected_ins_sets, self.cfg)
        self.configure_cpu()

    def configure_cpu(self):
        assert self.cfg is not None
        if isinstance(self.cpu, UserModeCPU) and self.cfg.stack_size != 0:
            self.cpu.setup_stack(self.cfg.stack_size)

    def register_all_arguments(self, parser: argparse.ArgumentParser):
        parser.add_argument(
            "files",
            metavar="file.asm",
            type=str,
            nargs="+",
            help="The assembly files to load, the last one will be run",
        )

        parser.add_argument(
            "--options",
            "-o",
            action=OptionStringAction,
            keys=(
                "disable_debug",
                "no_syscall_symbols",
                "fail_on_ex",
                "add_accept_imm",
                "unlimited_regs",
                "libc",
                "ignore_exit_code",
            ),
            help="""Toggle options. Available options are:
        disable_debug:        Disable ebreak instructions
        no_syscall_symbols:   Don't add symbols for SCALL_EXIT and others
        fail_on_ex:           If set, exceptions won't trigger the debugger
        add_accept_imm:       Accept "add rd, rs, imm" instruction (instead of addi)
        unlimited_regs:       Allow an unlimited number of registers
        libc:                 Load a libc-like runtime (for malloc, etc.)
        ignore_exit_code:     Don't exit with the programs exit code.""",
        )

        parser.add_argument(
            "--syscall-opts",
            "-so",
            action=OptionStringAction,
            keys=("fs_access", "disable_input"),
        )

        parser.add_argument(
            "--instruction-sets",
            "-is",
            action=OptionStringAction,
            help="Instruction sets to load, available are: {}. All are enabled by default".format(
                ", ".join(self.available_ins_sets)
            ),
            keys={k: True for k in self.available_ins_sets},
            omit_empty=True,
        )

        parser.add_argument(
            "--stack_size",
            type=int,
            help="Stack size of loaded programs, defaults to 8MB",
            nargs="?",
        )

        parser.add_argument(
            "--flen",
            type=int,
            help="hardware FLEN, either 32 or 64. Defaults to 64",
            nargs="?",
            default=64,
        )

        parser.add_argument(
            "-v",
            "--verbose",
            help="Verbosity level (can be used multiple times)",
            action="count",
            default=0,
        )

        parser.add_argument(
            "--interactive",
            help="Launch the interactive debugger instantly instead of loading any "
            "programs",
            action="store_true",
        )

        parser.add_argument(
            "--ignore-exit-code",
            help="Ignore exit code of the program and always return 0 if the program ran to completion.",
            action="store_true",
            default=False,
        )

    def register_all_isas(self):
        self.available_ins_sets.update(InstructionSetDict)

    def register_all_program_loaders(self):
        self.available_file_loaders.append(AssemblyFileLoader)

    def parse_argv(self, argv: List[str]):
        parser = argparse.ArgumentParser(
            description="RISC-V Userspace emulator",
            prog="riscemu",
            formatter_class=argparse.RawTextHelpFormatter,
        )
        if "--version" in argv:
            print(
                "riscemu version {}\n{}\n\nAvailable ISA: {}".format(
                    __version__, __copyright__, ", ".join(self.available_ins_sets)
                )
            )
            sys.exit()

        self.register_all_arguments(parser)

        # parse argv
        args = parser.parse_args(argv)

        # add ins
        if not hasattr(args, "ins"):
            setattr(args, "ins", {k: True for k in self.available_ins_sets})

        if args.flen not in (32, 64):
            raise ValueError("Cannot have flen other than 32 or 64!")

        # create RunConfig
        self.cfg = self.create_config(args)

        # set input files
        self.input_files.extend(args.files)

        # get selected ins sets
        self.selected_ins_sets.extend(
            self.available_ins_sets[name]
            for name, selected in args.ins.items()
            if selected
        )

        # remove floating point isas if they do not meet the flen requirements
        for isa in tuple(self.selected_ins_sets):
            if issubclass(isa, FloatArithBase) and isa.flen > args.flen:
                self.selected_ins_sets.remove(isa)

        # if "use_libc" is given, attach libc to path
        if self.cfg.use_libc:
            self.add_libc_to_input_files()

    def add_libc_to_input_files(self):
        """
        This adds the provided riscemu libc to the programs runtime.
        """
        for file in importlib_resources.files("riscemu.libc").iterdir():
            if file.name.lower().endswith(".s"):
                source = RiscemuSource(file.name, file.open("r"))
                self.input_files.append(source)

    def create_config(self, args: argparse.Namespace) -> RunConfig:
        # create a RunConfig from the cli args
        cfg_dict = dict(
            stack_size=args.stack_size,
            debug_instruction=not args.options["disable_debug"],
            include_scall_symbols=not args.options["no_syscall_symbols"],
            debug_on_exception=not args.options["fail_on_ex"],
            add_accept_imm=args.options["add_accept_imm"],
            unlimited_registers=args.options["unlimited_regs"],
            scall_fs=args.syscall_opts["fs_access"],
            scall_input=not args.syscall_opts["disable_input"],
            verbosity=args.verbose,
            use_libc=args.options["libc"],
            ignore_exit_code=args.options["ignore_exit_code"],
            flen=args.flen,
        )
        for k, v in dict(cfg_dict).items():
            if v is None:
                del cfg_dict[k]

        return RunConfig(**cfg_dict)

    def load_programs(self):
        for path in self.input_files:
            max_bid = -1
            bidder = None
            # get best-fit loader:
            for loader in self.available_file_loaders:
                if isinstance(path, RiscemuSource):
                    score = path.get_score_for(loader)
                else:
                    score = loader.can_parse(path)

                if score > max_bid:
                    max_bid = score
                    bidder = loader
            if max_bid <= 0:
                raise RuntimeError(
                    f"Cannot load {path}! No loader for this file type available."
                )
            if isinstance(path, RiscemuSource):
                stream = path.stream
                source_name = path.name
            elif path == "-":
                stream: IOBase = sys.stdin
                source_name = "<stdin>"
            else:
                source_name = path
                stream: IOBase = open(path, "rb" if bidder.is_binary else "r")

            programs = bidder.instantiate(source_name, stream, {}).parse()
            if isinstance(programs, Program):
                programs = [programs]
            for p in programs:
                self.cpu.mmu.load_program(p)
            if self.cfg.verbosity > 2:
                print(
                    FMT_GRAY
                    + "[Startup] Loaded {} with loader {}".format(
                        source_name, bidder.__name__
                    )
                    + FMT_NONE
                )

    def run_from_cli(self, argv: List[str]):
        # register everything
        self.register_all_isas()
        self.register_all_program_loaders()

        # parse argv and set up cpu
        self.parse_argv(argv)
        self.instantiate_cpu()
        self.load_programs()

        if self.cfg.verbosity > 3:
            print(
                FMT_GRAY
                + "[Startup] Startup complete, the following sections were loaded"
            )
            for sec in self.cpu.mmu.sections:
                print("          {}".format(sec.debug_str()))
            print(FMT_NONE)
        # run the program
        self.cpu.launch(self.cfg.verbosity > 1)

    def run(self):
        """
        This assumes that these values were set externally:

         - available_file_loaders: A list of available file loaders.
           Can be set using .register_all_program_loaders()
         - cfg: The RunConfig object. Can be directly assigned to the attribute
         - input_files: A list of assembly files to load.
         - selected_ins_sets: A list of instruction sets the CPU should support.
        """
        assert self.cfg is not None, "self.cfg must be set before calling run()"
        assert self.selected_ins_sets, "self.selected_ins_sets cannot be empty"
        assert self.input_files, "self.input_files cannot be empty"

        if self.cfg.use_libc:
            self.add_libc_to_input_files()

        self.instantiate_cpu()
        self.load_programs()

        # run the program
        self.cpu.launch(self.cfg.verbosity > 1)


class OptionStringAction(argparse.Action):
    def __init__(self, option_strings, dest, keys=None, omit_empty=False, **kwargs):
        if keys is None:
            raise ValueError('must define "keys" argument')
        if isinstance(keys, dict):
            keys_d = keys
        elif isinstance(keys, (list, tuple)):
            keys_d = {}
            for k in keys:
                if isinstance(k, tuple):
                    k, v = k
                else:
                    v = False
                keys_d[k] = v
        else:
            keys_d = dict()
        super().__init__(option_strings, dest, default=keys_d, **kwargs)
        self.keys = keys_d
        self.omit_empty = omit_empty

    def __call__(self, parser, namespace, values, option_string=None):
        d = {}
        if not self.omit_empty:
            d.update(self.keys)
        for x in values.split(","):
            if x in self.keys:
                d[x] = True
            else:
                raise ValueError("Invalid parameter supplied: " + x)
        setattr(namespace, self.dest, d)



"""
RiscEmu (c) 2021 Anton Lydike

SPDX-License-Identifier: MIT

This file holds the logic for starting the emulator from the CLI
"""
import sys

from .core import RiscemuBaseException
from .riscemu_main import RiscemuMain


def main():
    try:
        main = RiscemuMain()
        main.run_from_cli(sys.argv[1:])
        sys.exit(main.cpu.exit_code if not main.cfg.ignore_exit_code else 0)

    except RiscemuBaseException as e:
        print("Error: {}".format(e.message()))
        e.print_stacktrace()

        sys.exit(-1)


if __name__ == "__main__":
    main()



"""
Loads a memory image with debug information into memory
"""

import os.path
from io import IOBase
from typing import List, Iterable

from .ElfLoader import ElfMemorySection
from .types import MemoryImageDebugInfos
from ..assembler import INSTRUCTION_SECTION_NAMES
from ..colors import FMT_NONE, FMT_PARSE
from ..core import MemoryFlags, ProgramLoader, Program, T_ParserOpts


class MemoryImageLoader(ProgramLoader):
    is_binary = True

    @classmethod
    def can_parse(cls, source_name: str) -> float:
        if source_name.split(".")[-1] == "img":
            return 1
        return 0

    @classmethod
    def get_options(cls, argv: List[str]) -> [List[str], T_ParserOpts]:
        return argv, {}

    def parse(self) -> Iterable[Program]:
        if "debug" not in self.options:
            yield self.parse_no_debug()
            return

        with open(self.options.get("debug"), "r") as debug_file:
            debug_info = MemoryImageDebugInfos.load(debug_file.read())

        with self.source as source_file:
            data: bytearray = bytearray(source_file.read())

        for name, sections in debug_info.sections.items():
            program = Program(name)

            for sec_name, (start, size) in sections.items():
                if program.base is None:
                    program.base = start

                # in_code_sec = get_section_base_name(sec_name) in INSTRUCTION_SECTION_NAMES
                program.add_section(
                    ElfMemorySection(
                        data[start : start + size],
                        sec_name,
                        program.context,
                        name,
                        start,
                        MemoryFlags(False, True),
                    )
                )

            program.context.labels.update(debug_info.symbols.get(name, dict()))
            program.global_labels.update(debug_info.globals.get(name, set()))

            yield program

    def parse_no_debug(self) -> Program:
        print(
            FMT_PARSE
            + "[MemoryImageLoader] Warning: loading memory image without debug information!"
            + FMT_NONE
        )

        with self.source as source_file:
            data: bytes = source_file.read()

        p = Program(self.filename)
        p.add_section(
            ElfMemorySection(
                bytearray(data), ".text", p.context, p.name, 0, MemoryFlags(False, True)
            )
        )
        return p

    @classmethod
    def instantiate(
        cls, source_name: str, source: IOBase, options: T_ParserOpts
    ) -> "ProgramLoader":
        if os.path.isfile(source_name + ".dbg"):
            return MemoryImageLoader(
                source_name, source, dict(**options, debug=source_name + ".dbg")
            )
        return MemoryImageLoader(source_name, source, options)



from riscemu import RunConfig
from riscemu.core import Program
from .PrivCPU import PrivCPU

import sys

if __name__ == "__main__":
    import argparse

    parser = argparse.ArgumentParser(
        description="RISC-V privileged architecture emulator", prog="riscemu"
    )

    parser.add_argument(
        "source",
        type=str,
        help="Compiled RISC-V ELF file or memory image containing compiled RISC-V ELF files",
        nargs="+",
    )
    parser.add_argument(
        "--debug-exceptions",
        help="Launch the interactive debugger when an exception is generated",
        action="store_true",
    )

    parser.add_argument(
        "-v",
        "--verbose",
        help="Verbosity level (can be used multiple times)",
        action="count",
        default=0,
    )

    parser.add_argument(
        "--slowdown",
        help="Slow down the emulated CPU clock by a factor",
        type=float,
        default=1,
    )

    args = parser.parse_args()

    cpu = PrivCPU(
        RunConfig(
            verbosity=args.verbose,
            debug_on_exception=args.debug_exceptions,
            slowdown=args.slowdown,
        )
    )

    for source_path in args.source:
        loader = max(
            (loader for loader in cpu.get_loaders()),
            key=lambda l: l.can_parse(source_path),
        )
        argv, opts = loader.get_options(sys.argv)
        program = loader.instantiate(source_path, opts).parse()
        if isinstance(program, Program):
            cpu.load_program(program)
        else:
            program_iter = program
            for program in program_iter:
                cpu.load_program(program)

    cpu.launch(verbose=args.verbose > 4)



"""
RiscEmu (c) 2021 Anton Lydike

SPDX-License-Identifier: MIT
"""

from ..instructions.RV32I import *
from riscemu.core.exceptions import INS_NOT_IMPLEMENTED
from riscemu.core.traps import *
from riscemu.core.privmodes import PrivModes
from ..colors import FMT_CPU, FMT_NONE
import typing

if typing.TYPE_CHECKING:
    from riscemu.priv.PrivCPU import PrivCPU


class PrivRV32I(RV32I):
    cpu: "PrivCPU"
    """
    This is an extension of RV32I, written for the PrivCPU class
    """

    def instruction_csrrw(self, ins: "Instruction"):
        rd, rs, csr_addr = self.parse_crs_ins(ins)
        old_val = None
        if rd != "zero":
            self.cpu.csr.assert_can_read(self.cpu.mode, csr_addr)
            old_val = self.cpu.csr.get(csr_addr)
        if rs != "zero":
            new_val = self.regs.get(rs)
            self.cpu.csr.assert_can_write(self.cpu.mode, csr_addr)
            self.cpu.csr.set(csr_addr, new_val)
        if old_val is not None:
            self.regs.set(rd, old_val)

    def instruction_csrrs(self, ins: "Instruction"):
        rd, rs, csr_addr = self.parse_crs_ins(ins)
        if rs != "zero":
            # oh no, this should not happen!
            INS_NOT_IMPLEMENTED(ins)
        if rd != "zero":
            self.cpu.csr.assert_can_read(self.cpu.mode, csr_addr)
            old_val = self.cpu.csr.get(csr_addr)
            self.regs.set(rd, old_val)

    def instruction_csrrc(self, ins: "Instruction"):
        INS_NOT_IMPLEMENTED(ins)

    def instruction_csrrsi(self, ins: "Instruction"):
        INS_NOT_IMPLEMENTED(ins)

    def instruction_csrrwi(self, ins: "Instruction"):
        ASSERT_LEN(ins.args, 3)
        rd, imm, addr = ins.get_reg(0), ins.get_imm(1), ins.get_imm(2)
        if rd != "zero":
            self.cpu.csr.assert_can_read(self.cpu.mode, addr)
            old_val = self.cpu.csr.get(addr)
            self.regs.set(rd, old_val)
        self.cpu.csr.assert_can_write(self.cpu.mode, addr)
        self.cpu.csr.set(addr, imm)

    def instruction_csrrci(self, ins: "Instruction"):
        INS_NOT_IMPLEMENTED(ins)

    def instruction_mret(self, ins: "Instruction"):
        if self.cpu.mode != PrivModes.MACHINE:
            print("MRET not inside machine level code!")
            raise IllegalInstructionTrap(ins)
        # restore mie
        mpie = self.cpu.csr.get_mstatus("mpie")
        self.cpu.csr.set_mstatus("mie", mpie)
        # restore priv
        mpp = self.cpu.csr.get_mstatus("mpp")
        self.cpu.mode = PrivModes(mpp)
        # restore pc
        mepc = self.cpu.csr.get("mepc")
        self.cpu.pc = (mepc - self.cpu.INS_XLEN).abs_value

        if self.cpu.conf.verbosity > 0:
            sec = self.mmu.get_sec_containing(mepc.value)
            if sec is not None:
                print(
                    FMT_CPU
                    + "[CPU] returning to mode {} in {} (0x{:x})".format(
                        PrivModes(mpp).name, self.mmu.translate_address(mepc), mepc
                    )
                    + FMT_NONE
                )
                if self.cpu.conf.verbosity > 1:
                    self.regs.dump_reg_a()

    def instruction_uret(self, ins: "Instruction"):
        raise IllegalInstructionTrap(ins)

    def instruction_sret(self, ins: "Instruction"):
        raise IllegalInstructionTrap(ins)

    def instruction_scall(self, ins: "Instruction"):
        """
        Overwrite the scall from userspace RV32I
        """
        raise EcallTrap(self.cpu.mode)

    def instruction_beq(self, ins: "Instruction"):
        rs1, rs2, dst = self.parse_rs_rs_imm(ins)
        if rs1 == rs2:
            self.pc += dst.abs_value - 4

    def instruction_bne(self, ins: "Instruction"):
        rs1, rs2, dst = self.parse_rs_rs_imm(ins)
        if rs1 != rs2:
            self.pc += dst.abs_value - 4

    def instruction_blt(self, ins: "Instruction"):
        rs1, rs2, dst = self.parse_rs_rs_imm(ins)
        if rs1 < rs2:
            self.pc += dst.abs_value - 4

    def instruction_bge(self, ins: "Instruction"):
        rs1, rs2, dst = self.parse_rs_rs_imm(ins)
        if rs1 >= rs2:
            self.pc += dst.abs_value - 4

    def instruction_bltu(self, ins: "Instruction"):
        rs1, rs2, dst = self.parse_rs_rs_imm(ins)
        if rs1.unsigned_value < rs2.unsigned_value:
            self.pc += dst.abs_value - 4

    def instruction_bgeu(self, ins: "Instruction"):
        rs1, rs2, dst = self.parse_rs_rs_imm(ins)
        if rs1.unsigned_value >= rs2.unsigned_value:
            self.pc += dst.abs_value - 4

    # technically deprecated
    def instruction_j(self, ins: "Instruction"):
        raise NotImplementedError("Should never be reached!")

    def instruction_jal(self, ins: "Instruction"):
        ASSERT_LEN(ins.args, 2)
        reg = ins.get_reg(0)
        addr = ins.get_imm(1)
        if reg == "ra" and (
            (self.cpu.mode == PrivModes.USER and self.cpu.conf.verbosity > 1)
            or (self.cpu.conf.verbosity > 3)
        ):
            print(
                FMT_CPU
                + "Jumping from 0x{:x} to {} (0x{:x})".format(
                    self.pc,
                    self.mmu.translate_address(self.pc + addr.pcrel_value),
                    self.pc + addr.pcrel_value,
                )
                + FMT_NONE
            )
            self.regs.dump_reg_a()
        self.regs.set(reg, Int32(self.pc))
        self.pc += addr.pcrel_value - 4

    def instruction_jalr(self, ins: "Instruction"):
        ASSERT_LEN(ins.args, 3)
        rd, rs, imm = self.parse_rd_rs_imm(ins)
        self.regs.set(rd, Int32(self.pc))
        self.pc = rs.value + imm.abs_value - 4

    def instruction_sbreak(self, ins: "Instruction"):
        raise LaunchDebuggerException()

    def parse_crs_ins(self, ins: "Instruction"):
        ASSERT_LEN(ins.args, 3)
        return ins.get_reg(0), ins.get_reg(1), ins.get_imm(2)

    def parse_mem_ins(self, ins: "Instruction") -> Tuple[str, int]:
        ASSERT_LEN(ins.args, 3)
        addr = self.regs.get(ins.get_reg(1)) + ins.get_imm(2).abs_value
        reg = ins.get_reg(0)
        return reg, addr
